import abc
import logging

from jsonschema import validate


class Transformer:
    @classmethod
    def match(cls, input, schema):
        try:
            validate(instance=input, schema=schema)
            return True
        except Exception as e:
            logging.info(e)
            return False

    @abc.abstractclassmethod
    def _original_schema(cls):
        # Schema to match against on input
        raise NotImplementedError("Transformer: Original Schema")

    @abc.abstractclassmethod
    def _normalized_schema(cls):
        # Schema to match against on output
        raise NotImplementedError("Transformer: Nomralized Schema")

    @abc.abstractclassmethod
    def _transform(cls, input):
        # Actual transform logic
        raise NotImplementedError

    @classmethod
    def _after_transform(cls, input):
        return input

    @classmethod
    def transform(cls, incoming_json):
        # if cls.match(incoming_json, cls._original_schema()) == False:
        #     raise Exception(
        #         (
        #             "Error: Can't transform, incoming JSON doesn't match"
        #             " schema.\n\nSchema: {}\n\nInput: {}"
        #         ),
        #         cls._original_schema(),
        #         incoming_json,
        #     )

        outgoing_json = cls._transform(incoming_json)
        outgoing_json = cls._after_transform(outgoing_json)

        # TODO: Check transform worked properly
        # if cls.match(outgoing_json, cls._normalized_schema()) == False:
        #     raise Exception("Error: Can't transform, outgoing JSON doesn't match schema.\n\nSchema: {}\n\nOutgoing: {}", cls._normalized_schema(), outgoing_json)

        return outgoing_json



from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from altimate_models.base.source import DataSource

from altimate_models.shared_models import Policy, ProfileConfig, Resource, ResourceType
from altimate_profiler.factory import (
    DialectFactory,
    MetadataFactory,
    ProfilerFactory,
    SchemaTransformerFactory,
)
from altimate_profiler.transformer.metrics import MetricTransformer


def get_schema_profile(
    data_source: "DataSource", profile_config: ProfileConfig, use_account_usage: bool = False
) -> List:
    schemas = []
    for policy in profile_config.policies:
        profiler = ProfilerFactory.create(data_source, policy, use_account_usage)
        discovered_data = profiler.get_discovery_data()
        transformer = SchemaTransformerFactory.create(data_source.type)
        schema = transformer.transform(discovered_data)
        schemas.append(schema)

    return schemas


def get_metrics_profile(
    data_source: "DataSource", profile_config: ProfileConfig, use_account_usage: bool = False
) -> List:
    metrics = []
    for policy in profile_config.policies:
        profiler = ProfilerFactory.create(data_source, policy, use_account_usage)

        raw_metrics = profiler.get_metrics_data()
        transformer = MetricTransformer()

        transformed_metrics = transformer.transform(raw_metrics)

        metrics.append(transformed_metrics)

    return metrics


def get_metadata(data_source: "DataSource", resource: Resource) -> List:
    dialect = DialectFactory.create(data_source)
    if resource.resource_type == ResourceType.FILE:
        resource.resource.options = dialect.get_options(resource.resource.options)
        resource_name = dialect.create_table_path(resource.resource, dialect.DATA_QUERY)
    else:
        resource_name = dialect.create_table_path(resource.resource)

    if resource.filters:
        resource_name = f"{resource_name} WHERE {resource.filters}"

    policy = Policy(
        policy_index=0, resources=[resource], terms=[], policy_type="metadata"
    )
    profiler = MetadataFactory.create(data_source, policy, resource_name)
    return profiler.get_metadata()


def get_debug_data(data_source: "DataSource", debug_sql: str) -> List:
    policy = Policy(policy_index=0, resources=[], terms=[], policy_type="metadata")
    profiler = MetadataFactory.create(data_source, policy, "")
    return profiler.get_debugdata(debug_sql)



import logging
import sys
from typing import Text

from altimate_models.constants import TEST_SQL_QUERY

try:
    import duckdb
except ImportError:
    pass

from altimate_profiler.exceptions import (
    INVALID_UNICODE_IN_CSV,
    AltimateInvalidCharactersInCSVException,
    AltimateInvalidInputException,
)


class DuckDBExtractor:
    def __init__(
        self,
        connection_str: Text,
    ):
        self.connection_str = connection_str
        self.engine = None

    def _initialize(self):
        raise NotImplementedError(
            "Cannot use DuckDBExtractor directly. Please use an implementation like S3Extractor."
        )

    def _create_engine(self, connection_str: Text):
        if not "duckdb" in sys.modules:
            raise ImportError("Could not import duckdb. Please install duckdb.")
        engine = duckdb.connect(database=":memory:", read_only=False)
        engine.execute(connection_str)
        return engine

    def _execute(self, sql: Text):
        if not self.engine:
            raise Exception("Connection has already been closed. Could not execute.")
        try:
            cs = self.engine.sql(sql)
            results = self._retrieve_results(cs)
        except duckdb.InvalidInputException as e:
            if INVALID_UNICODE_IN_CSV in str(e):
                raise AltimateInvalidCharactersInCSVException(
                    "CSV file contains invalid unicode characters. This is not supported currently."
                    "We only support UTF-8 encoded / ISO-8849 CSV files."
                )
            else:
                raise AltimateInvalidInputException(
                    f"The column configuration might be incorrect." f" Exception: {e}"
                )

        return results

    def _retrieve_results(self, cs):
        data_dict = cs.fetchdf().to_dict("split")
        columns = data_dict["columns"]
        rows = [dict(zip(columns, row)) for row in data_dict["data"]]

        return {
            "columns": columns,
            "rows": rows,
        }

    def _terminate(self):
        if self.engine:
            self.engine.close()
            self.engine = None

    def test(self):
        try:
            self._initialize()
            result = self._execute(TEST_SQL_QUERY)
            self._terminate()

            rows = result["rows"]
            columns = result["columns"]

            return len(rows) == 1 and rows[0][columns[0]] == 1  # TODO: Fix Indexing
        except Exception as e:
            logging.error(e)
            return False

    def run(self, sql: Text):
        self._initialize()

        results = self._execute(sql)

        self._terminate()

        return results



from typing import List, Text

from altimate_profiler.builders.base_builder import Base
from altimate_profiler.metrics.metadata_metric_type import MetadataMetricType


class MetadataQueryBuilder(Base):
    def __init__(self, dialect):
        self.dialect = dialect

    def _gen_sql(self, column, metric):
        attr = getattr(self.dialect, metric._value_)
        return f"{attr().format(column=column)} AS {column}__{metric.name}"

    def get_metrics(
        self,
        columns: List[Text],
        resource_name: Text,
        metrics: List[MetadataMetricType],
    ) -> Text:
        queries = []
        for column in columns:
            for metric in metrics:
                queries.append(self._gen_sql(column, metric))
        return f"SELECT {', '.join(queries)} FROM {resource_name}"

    def row_count(self, resource_name: Text) -> Text:
        return f"SELECT COUNT(*) FROM {resource_name}"

    def compile(self) -> List[Text]:
        metrics_queries = []
        for term in self.policy.terms:
            metrics_queries.append(
                self._gen_sql(MetadataMetricType(term.metric), term.params)
            )



from enum import Enum


class ColumnMetricType(Enum):
    APPROX_DISTINCTNESS = "_approx_distinctness"
    COMPLETENESS = "_completeness"
    ZERO_RATE = "_zero_rate"
    NEGATIVE_RATE = "_negative_rate"
    MEAN = "_numeric_mean"
    MIN = "_numeric_min"
    MAX = "_numeric_max"
    NUMERIC_STD = "_numeric_std"
    APPROX_DISTINCT_COUNT = "_approx_distinct_count"
    MEAN_LENGTH = "_mean_length"
    MAX_LENGTH = "_max_length"
    MIN_LENGTH = "_min_length"
    STD_LENGTH = "_std_length"
    TEXT_INT_RATE = "_text_int_rate"
    TEXT_NUMBER_RATE = "_text_number_rate"
    TEXT_UUID_RATE = "_text_uuid_rate"
    TEXT_ALL_SPACES_RATE = "_text_all_spaces_rate"
    TEXT_NULL_KEYWORD_RATE = "_text_null_keyword_rate"
    CATEGORIES = "_categories"
    FUTURE_DATES_RATE = "_future_dates_rate"
    INVALID_EMAIL_RATE = "_invalid_email_rate"
    PERCENTAGE_SELECTS = "_percentage_selects"
    PERCENTAGE_INSERTS = "_percentage_inserts"
    PERCENTAGE_WAREHOUSE_SIZE_X_SMALL = "_percentage_warehouse_size_x_small"
    PERCENTAGE_FAILURES = "_percentage_failures"
    AVERAGE_NON_NEGATIVE = "_average_non_negative"
    MAX_NON_NEGATIVE = "_max_non_negative"
    MIN_NON_NEGATIVE = "_min_non_negative"
    PERCENTAGE_WAREHOUSE_IS_XSMALL = "_percentage_warehouse_is_xsmall"
    AVG_ROWS_INSERTED_OR_UPDATED = "_avg_rows_inserted_or_updated"
    AVG_SELECT_QUERY_TIME = "_avg_select_query_time"
    CONDITIONAL_MAX = "_conditional_max"
    CONDITIONAL_MIN = "_conditional_min"
    CONDITIONAL_MEAN = "_conditional_mean"
    CONDITIONAL_COUNT = "_conditional_count"
    CONDITIONAL_PERCENTAGE = "_conditional_percentage"
    NULL_COUNT = "_null_count"
    NULL_PERCENTAGE = "_null_percentage"
    CARDINALITY = "_cardinality"
    CATEGORY_COUNT = "_cardinality"
    REGEX_MATCH_PERCENTAGE = "_regex_match_percentage"
    LIKE_PERCENTAGE = "_like_percentage"
    DATES_IN_FUTURE_PERCENTAGE = "_dates_in_future_percentage"
    ALLOWED_VALUES = "_allowed_values"
    NUMERIC_MODE = "_numeric_mode"
    NUMERIC_MEDIAN = "_numeric_median"
    PERCENTILE = "_percentile"
    NUMERIC_VARIANCE = "_numeric_variance"



import types
from contextlib import contextmanager
from typing import Any, Dict, Generator, List, Optional, Text, Type

import sshtunnel

from altimate_models.base.source import DataSource
from altimate_models.constants import TEST_SQL_QUERY
from altimate_profiler.exceptions import AltimateDataStoreConnectionException
from altimate_profiler.utils import green_bold_string, red

# SUPPORTED_DATASOURCE_TYPES = Type[PostgresSource]


class SSHSource(DataSource):
    # this class should only be initialized if we have a jump host to worry about
    # if we don't have a jump host, then we should just use the original datasource
    # which is why this is not an Optional field
    jump_info: dict
    # in general, this should be one of the implementations
    # of the abstract class DataSource. currently only supports
    # postgres. However, since sshsource gets initialized in a factory
    # with abstract datastore type so we have to type it as such or else
    # pydantic shouts at us
    original_ds: DataSource
    available_delegated_methods: Optional[List] = None
    available_delegated_props: Optional[List] = None

    def __init__(self, **data):
        # A lot seems to depend on name and type.
        # changing these seems to break stuff so keeping them the same.
        # Ideally, we should have a prefix when using the delegated datasource.
        # maybe a future TODO.
        data["type"] = data[
            "original_ds"
        ].type  # "ssh::{}".format(data["original_ds"].type)
        data["name"] = data[
            "original_ds"
        ].name  # "ssh::{}".format(data["original_ds"].name)
        super().__init__(**data)

        # populate delegated method list
        # only allow delegation to public methods
        # special case made for properties because we are
        # also delegating all that to original ds
        self.available_delegated_methods = [
            f
            for f in dir(self.original_ds)
            if (
                not f.startswith("_")
                and isinstance(getattr(self.original_ds, f), types.MethodType)
            )
        ]
        self.available_delegated_props = [
            f
            for f in dir(self.original_ds)
            if (not f.startswith("_") and f not in self.available_delegated_methods)
        ]

    # DO NOT REMOVE: we go through __getattr__ only if methods do not exist in this or parent class.
    # but because we want sshsource to also mimick a datastore, we need to explicitly override these
    # delegated method to make sure they works.
    def get_connection_string(self, *args, **kwargs) -> Text:
        return self.original_ds.get_connection_string(*args, **kwargs)

    def get_resource(self):
        return self.original_ds.get_resource()

    def get_resource_config(self, resource_config: Dict):
        return self.original_ds.get_resource_config(resource_config)

    def test_connection(self):
        # need to redo all public facing endpoints it seems
        # return self.original_ds.test_connection()
        try:
            with self.get_engine() as engine:
                with engine.connect() as connection:
                    connection.execute(TEST_SQL_QUERY)
                print(
                    green_bold_string(
                        f"Successfully connected to {self.original_ds.name} via jumphost!"
                    )
                )

        except Exception as e:
            print(red("Connection via jump host failed."))
            print("Check your credentials and try again.")
            raise AltimateDataStoreConnectionException(
                f"Connection to {self.original_ds.name} failed."
                f" Please check the credentials"
            )

    # overriding the drop_credentials method to avoid dropping the jump_info
    # thats not held in the original ds
    def drop_credentials(self):
        conf_dict = self.original_ds.drop_credentials()
        if self.jump_info:
            conf_dict["jump_info"] = self.jump_info

        return conf_dict

    def get_identifier_and_fqn(self):
        return self.original_ds.get_identifier_and_fqn()

    def update_credentials(self, key: str):
        return self.original_ds.update_credentials(key)

    def override_schema(self, schema):
        return self.original_ds.override_schema(schema)

    def get_connect_args(self):
        return self.original_ds.get_connect_args()

    # delegator to the original datasource object. This should catch any methods not
    # overridden above (baseclass does not have specific methods that are part of
    # pg or snowflake implementations eg. host/port or warehouse etc)
    # and use the original datasource to generate responses
    def __getattr__(self, prop_or_func):
        # we are also delegating properties so we need to handle them separately
        # these are not functions but actual values so instead of the method wrapper
        # for functions below, these will be returned immediately
        if prop_or_func in self.available_delegated_props:
            # print("delegating property: {}".format(prop_or_func))
            value = getattr(self.original_ds, prop_or_func)
            return value

        # if the property is not a value, then check if its in the list of methods
        if prop_or_func in self.available_delegated_methods:
            # print("delegate method {} to original ds".format(prop_or_func))

            def method(*args):
                return getattr(self.original_ds, prop_or_func)(*args)

            return method

        else:
            raise AttributeError

    @contextmanager
    def get_bound_host_and_port(self) -> Generator[Dict[str, Any], None, None]:
        sshtunnel.DEFAULT_LOGLEVEL = sshtunnel.logging.DEBUG
        # allowing "SSH_PKEY_PW" to not exist because sometimes it private key can be passwordless
        if not all(
            k in self.jump_info for k in ["SSH_PKEY_PATH", "SSH_HOST", "SSH_USER"]
        ):
            raise ValueError("Missing required keys in jump_info")

        with sshtunnel.open_tunnel(
            (
                self.jump_info["SSH_HOST"],
                int(22),
            ),  # defaulting to bind port of https
            ssh_username=self.jump_info["SSH_USER"],
            ssh_pkey=self.jump_info["SSH_PKEY_PATH"],
            ssh_private_key_password=self.jump_info.get("SSH_PKEY_PW", None),
            remote_bind_address=(self.host, int(self.port)),  # type: ignore
        ) as tunnel:
            try:
                yield {"host": "0.0.0.0", "port": tunnel.local_bind_port}  # type: ignore
            except Exception as e:
                # re-raise exceptions raised in the context block
                raise e



import sys
from typing import Dict, Optional, Text

try:
    import click
    from sqlalchemy import create_engine
except ImportError:
    pass

from altimate_models.base.source import DataSource
from altimate_models.shared_models import Resource, ResourceType, TableResource
from altimate_profiler.exceptions import AltimateDataStoreConnectionException
from altimate_profiler.utils import green_bold_string, red, sha256_encoding


class PostgresSource(DataSource):
    NAME = "postgres"
    host: Optional[Text]
    port: Optional[Text]
    user: Optional[Text]
    password: Optional[Text]
    database: Optional[Text]
    sync_credentials: bool = False

    def get_connection_string(self, *args, **kwargs) -> Text:
        host = kwargs.get("host", self.host)
        port = kwargs.get("port", self.port)
        return "postgresql://{user}:{password}@{host}:{port}/{database}".format(
            user=self.user,
            password=self.password,
            host=host,
            port=port,
            database=self.database,
        )

    @classmethod
    def get_connection_information(cls, data_store_config: Dict):
        data_store_config[
            "connection_info"
        ] = f"{data_store_config['host']}:{data_store_config['port']}/{data_store_config['database']}"
        data_store_config["connection_hash"] = sha256_encoding(
            data_store_config["connection_info"]
        )
        return data_store_config

    @classmethod
    def prompt(cls, data_store_config: Dict):
        if "click" not in sys.modules:
            raise ImportError(
                "click is not installed. Please install click to connect to S3."
            )
        data_store_config["host"] = click.prompt("host", type=str)
        data_store_config["port"] = click.prompt("port", type=int)
        data_store_config["user"] = click.prompt("user", type=str)
        data_store_config["password"] = click.prompt(
            "password", type=str, hide_input=True
        )
        data_store_config["database"] = click.prompt("database", type=str)
        data_store_config["sync_credentials"] = click.confirm(
            "sync credentials with altimate?", default=False
        )
        data_store_config = cls.get_connection_information(data_store_config)
        return PostgresSource(**data_store_config)

    def cred_prompt(self):
        if "click" not in sys.modules:
            raise ImportError(
                "click is not installed. Please install click to connect to S3."
            )
        self.user = click.prompt("user", type=str)
        self.password = click.prompt("password", type=str, hide_input=True)
        self.sync_credentials = click.confirm(
            "sync credentials with altimate?", default=False
        )
        conn_info = self.get_connection_information(self.__dict__)
        self.connection_info = conn_info["connection_info"]
        self.connection_hash = conn_info["connection_hash"]

    def test_connection(self):
        if not "sqlalchemy" in sys.modules:
            raise ImportError(
                "sqlalchemy is not installed. Please install sqlalchemy to test connections"
            )
        try:
            with self.get_engine() as engine:
                engine.connect()
                print(green_bold_string("Successfully connected to Postgres!"))

        except Exception as e:
            print(red("Connection to Postgres failed."))
            print("Check your credentials and try again.")
            raise AltimateDataStoreConnectionException(
                f"Connection to {self.name} failed." f" Please check the credentials"
            )

    def drop_credentials(self):
        config_dict = self.dict()
        password = config_dict.pop("password")
        username = config_dict.pop("user")
        if self.sync_credentials:
            config_dict["key"] = f"altimate:{username}:{password}"
        return config_dict

    def get_resource(self):
        return Resource(
            resource_type=ResourceType.TABLE,
            resource=TableResource(
                database=self.database,
            ),
        )

    def get_identifier_and_fqn(self):
        if "click" not in sys.modules:
            raise ImportError(
                "click is not installed. Please install click to connect to S3."
            )
        schema = click.prompt("schema", type=str)
        table = click.prompt("table", type=str)
        identifier = f"{self.NAME}://{self.database}.{schema}.{table}"
        fqn = f"database={self.database}/schema={schema}/table={table}"
        return identifier, fqn

    @classmethod
    def get_resource_config(cls, resource_config: Dict):
        filters = resource_config.get("filters")
        identifier = resource_config.get("identifier")
        data_store_type, path = identifier.split("://")
        database, schema, table = path.split(".")
        resource = TableResource(database=database, schema=schema, table=table)
        return Resource(
            resource_type=ResourceType.TABLE, resource=resource, filters=filters
        )

    @classmethod
    def map_config_to_source(cls, config: Dict):
        connection_info = config.get("connection_info")
        if not connection_info:
            return config
        postgres_config = cls.get_config_from_url(connection_info)
        config["host"] = postgres_config["host"]
        config["port"] = postgres_config["port"]
        config["database"] = postgres_config["database"]
        config.pop("connection_info")
        return config

    @classmethod
    def get_config_from_url(cls, connection_info):
        url, database = connection_info.split("/")
        host, port = url.split(":")
        return {"host": host, "port": port, "database": database}

    def update_credentials(self, credentials: str):
        _user, _pass = credentials.split(":")
        self.user = _user
        self.password = _pass

    def override_schema(self, schema):
        return schema

    def get_connect_args(self):
        connect_args = super().get_connect_args()
        if self.statement_timeout_in_ms is not None:
            connect_args[
                "options"
            ] = f"-c statement_timeout={self.statement_timeout_in_ms}"
        return connect_args



from typing import TYPE_CHECKING

from altimate_models.base.extractor import SQLAlchemyExtractor
from altimate_models.base.profiler import SQLAlchemyProfiler
from altimate_models.shared_models import Policy
from altimate_models.snowflake.dialect import SnowflakeSourceDialect

if TYPE_CHECKING:
    from altimate_models.snowflake.source import SnowflakeSource


class SnowflakeProfiler(SQLAlchemyProfiler):
    def __init__(self, data_source: "SnowflakeSource", policy: Policy, use_account_usage: bool = False):
        # FIXME - raise exception if the database in datasource
        # and that in policy is not the same.
        # data_source.database = policy.resource0.resource.database
        super().__init__(
            extractor=SQLAlchemyExtractor(
                connection_str=data_source.get_connection_string(
                    # database=policy.resource0.resource.database,
                    # schema=policy.resource0.resource.db_schema,
                ),
                data_source=data_source,
            ),
            policy=policy,
            dialect=SnowflakeSourceDialect(),
            use_account_usage=use_account_usage,
        )



import hashlib
from typing import Text


def sha256_encoding(string):
    sha256 = hashlib.sha256()
    sha256.update(string.encode("utf-8"))
    return sha256.hexdigest()


# Print utils
def green_bold_string(s: Text) -> Text:
    return f"[bold green]{s}[/bold green]"


def yellow(s: Text) -> Text:
    return f"[yellow]{s}[/yellow]"


def red(s: Text) -> Text:
    return f"[red]{s}[/red]"



from typing import List, Text

from altimate_models.shared_models import Policy
from altimate_profiler.builders.base_builder import Base


class CustomQueryBuilder(Base):
    def __init__(self, policy: Policy):
        self.policy = policy

    def compile(self) -> List[Text]:
        metrics_queries = []
        for term in self.policy.terms:
            # TODO: Add support for other metrics. Create a Dialect
            if term.params.get("sql"):
                metrics_queries.append(term.params.get("sql"))

        return metrics_queries



from typing import List, Text

from altimate_profiler.exceptions import AltimateProfilerException
from altimate_profiler.builders.s3_table_query_builder import S3TableQueryBuilder
from altimate_models.shared_models import Policy, S3Resource
from altimate_profiler.builders.s3_column_metrics_query_builder import S3ColumnMetricsQueryBuilder


class DuckDBSourceDialect:

    @classmethod
    def create_table_path(cls, resource: S3Resource, query_type: Text) -> Text:
        raise NotImplementedError

    @classmethod
    def _row_count(cls, table, params):
        return "select COALESCE(COUNT(*), 0)  from {table}".format(table=table)

    @classmethod
    def _duplicate_count(cls, table, params):
        return (
            "select COALESCE((SUM(c) - COUNT(c)), 0) as dulicates from (select md5(row_to_json(ROW({columns}))::TEXT),count(*) as c   from {table} group by md5(row_to_json(ROW({columns}))::TEXT) having count(*) > 1) as t"
            .format(columns=",".join(params["columns"]), table=table)
        )

    @classmethod
    def _percentage_distinct_count(cls, table, params):
        concat_columns = "CONCAT(" + ",".join([f"coalesce({column}:: text, '')" for column in params.get("columns", [])]) + ")"
        sql = """        
        WITH total_rows AS (
          SELECT COUNT(*) AS total
          FROM {table}
        ),
        distinct_rows AS (
          SELECT COUNT(DISTINCT {concat_columns}) AS num_distinct
          FROM {table}
        )
        SELECT
          ROUND((num_distinct * 100.0) / total, 2) AS percentage_distinct_rows
        FROM total_rows, distinct_rows;

        """
        return sql.format(concat_columns=concat_columns, table=table)


    @classmethod
    def _approx_distinct_count(cls):
        return "COUNT(DISTINCT {column})"

    @classmethod
    def _approx_distinctness(cls):
        return "{} / CAST(COUNT(*) AS NUMERIC)".format(cls._approx_distinct_count())

    @classmethod
    def _freshness(cls, table, params):
        time_unit = params.get("timeUnit", "day")
        column = params.get("column")
        if column is None:
            raise AltimateProfilerException("Column is required for freshness")
        return "select MIN(date_diff('{timeUnit}', {column}:: DATE, current_date)) from {table}".format(timeUnit=time_unit, column=column, table=table)

    @classmethod
    def _numeric_mean(cls):
        return "AVG({column})"

    @classmethod
    def _numeric_min(cls):
        return "MIN({column})"

    @classmethod
    def _numeric_max(cls):
        return "MAX({column})"

    @classmethod
    def _numeric_std(cls):
        return "stddev_pop(CAST({column} as double))"

    @classmethod
    def _mean_length(cls):
        return "AVG(LENGTH({column}))"

    @classmethod
    def _max_length(cls):
        return "MAX(LENGTH({column}))"

    @classmethod
    def _min_length(cls):
        return "MIN(LENGTH({column}))"

    @classmethod
    def _dates_in_future_percentage(cls):
        return "SUM(CASE WHEN {column} > current_date THEN 1 ELSE 0 END) * 100 / COALESCE(COUNT(*), 1)"


    @classmethod
    def _std_length(cls):
        return "STDDEV(CAST(LENGTH({column}) as double))"

    @classmethod
    def _categories(cls):
        return "ARRAY_AGG(DISTINCT CAST({column} AS varchar))"

    @classmethod
    def _allowed_values(cls):
        return "SUM(CASE WHEN {column} NOT IN {values} THEN 1 ELSE 0 END)"

    @classmethod
    def _num_categories(cls):
        return "COUNT(DISTINCT CAST({column} AS varchar))"

    @classmethod
    def _conditional_max(cls):
        return "MAX(CASE WHEN {condition} THEN {column} ELSE NULL END)"

    @classmethod
    def _conditional_min(cls):
        return "MIN(CASE WHEN {condition} THEN {column} ELSE NULL END)"

    @classmethod
    def _conditional_percentage(cls):
        return "SUM(CASE WHEN {condition} THEN 1 ELSE 0 END) * 100 / COALESCE(COUNT(*), 1)"

    @classmethod
    def _conditional_mean(cls):
        return "AVG(CASE WHEN {condition} THEN {column} ELSE NULL END)"

    @classmethod
    def _conditional_count(cls):
        return "SUM(CASE WHEN {condition} THEN 1 ELSE 0 END)"

    @classmethod
    def _null_count_table(cls, table, params):
        condition = " AND ".join(
            [f"{column} IS NULL" for column in params["columns"]]
        )
        return f"SELECT SUM(CASE WHEN {condition} THEN 1 ELSE 0 END) FROM {table}".format(table=table,
                                                                                          condition=condition)

    @classmethod
    def _null_percentage(cls):
        return "SUM(CASE WHEN {column} IS NULL THEN 1 ELSE 0 END) *100 / COALESCE(COUNT(*), 1)"

    @classmethod
    def _null_count(cls):
        return "SUM(CASE WHEN {column} IS NULL THEN 1 ELSE 0 END)"

    @classmethod
    def _cardinality(cls):
        return "COUNT(DISTINCT md5({column}))"

    @classmethod
    def _regex_match_percentage(cls):
        return "SUM(CASE WHEN regexp_full_match({column}, '{pattern}') THEN 1.0 ELSE 0.0 END) * 100 / COALESCE(COUNT(*), 1)"

    @classmethod
    def _like_percentage(cls):
        return "SUM(CASE WHEN {column} like '{pattern}' THEN 1.0 ELSE 0.0 END) * 100 / COALESCE(COUNT(*), 1)"

    @classmethod
    def _numeric_mode(cls):
        return "MODE({column})"

    @classmethod
    def _numeric_variance(cls):
        return "var_pop({column})"

    @classmethod
    def _percentile(cls):
        return "PERCENTILE_CONT({value}) within group (order by {column})"

    @classmethod
    def _numeric_median(cls):
        return "median({column})"

    @classmethod
    def _text_int_rate(cls):
        raise NotImplementedError

    @classmethod
    def _text_number_rate(cls):
        raise NotImplementedError

    @classmethod
    def _text_uuid_rate(cls):
        raise NotImplementedError

    @classmethod
    def text_all_spaces_rate(cls):
        raise NotImplementedError

    @classmethod
    def _text_null_keyword_rate(cls):
        raise NotImplementedError

    @classmethod
    def _zero_rate(cls):
        raise NotImplementedError

    @classmethod
    def _negative_rate(cls):
        raise NotImplementedError

    @classmethod
    def _completeness(cls):
        raise NotImplementedError

    @classmethod
    def schema_tables_query(cls, bucket: Text,  path: Text, options: Text):
        raise NotImplementedError

    @classmethod
    def schema_columns_query(cls, path: Text, options: Text):
        raise NotImplementedError

    @classmethod
    def column_metrics_query(cls,  policy: Policy) -> List[Text]:
        builder = S3ColumnMetricsQueryBuilder(cls,  policy)
        query = builder.compile()
        return query

    @classmethod
    def table_metrics_query(cls, policy: Policy) -> List[Text]:
        builder = S3TableQueryBuilder(cls,  policy)
        query = builder.compile()
        return query

    @classmethod
    def _base_query_sample(cls, select_sql: Text, bucket: Text, file_resource: S3Resource = None, filters: Text = None):
        raise NotImplementedError

    @classmethod
    def access_logs_query(cls):
        raise NotImplementedError

    @classmethod
    def copy_and_load_logs_query(cls):
        raise NotImplementedError



from altimate_models.base.extractor import SQLAlchemyExtractor


class PostgreSQLExtractor(SQLAlchemyExtractor):
    def _initialize(self):
        super()._initialize()

        try:
            self.connection.execution_options(
                postgresql_readonly=True,
            )
        except Exception:
            raise



from dataclasses import dataclass
from typing import Text, List

from altimate_models.duckdb.dialect import DuckDBSourceDialect
from altimate_models.duckdb.extractor import DuckDBExtractor
from altimate_profiler.builders.s3_schema_query_builder import S3SchemaQueryBuilder
from altimate_models.shared_models import Policy
from altimate_profiler.builders.custom_query_builder import CustomQueryBuilder


@dataclass
class DuckDBProfiler:
    extractor: DuckDBExtractor
    policy: Policy
    dialect: DuckDBSourceDialect

    def _columns_schema(self, _) -> Text:
        raise NotImplementedError

    def _tables_schema(self, _) -> Text:
        return self.dialect.schema_tables_query(
            path=self.policy.resource0.resource.path,
            options=self.policy.resource0.resource.options
        )

    def _access_logs(self, _) -> Text:
        return self.dialect.access_logs_query()

    def _copy_and_load_logs(self, _) -> Text:
        return self.dialect.copy_and_load_logs_query()

    def _metrics_query(self) -> List[Text]:
        queries = self.dialect.column_metrics_query(
            self.policy
        )
        return queries

    def _custom_query(self) -> List[Text]:
        custom_metric_builder = CustomQueryBuilder(self.policy)
        return custom_metric_builder.compile()

    def _table_query(self):
        return self.dialect.table_metrics_query(self.policy)

    def _schema_policy(self):
        discovery_data = self.get_discovery_data()
        result = S3SchemaQueryBuilder(self.policy, discovery_data).compile()
        return result

    def _discovery_query(self):
        return self.dialect.schema_tables_query(
            bucket=self.policy.resource0.resource.bucket,
            path=self.policy.resource0.resource.path,
            options=self.policy.resource0.resource.options
        )

    def get_discovery_data(self):
        schema = self.extractor.run(self._discovery_query())
        columns = schema["columns"]
        rows = schema["rows"]
        columns.append("path")
        columns.append("file_format")
        for row in rows:
            row["file_path"] = self.policy.resource0.resource.path
            row["file_format"] = self.policy.resource0.resource.format
            row["bucket"] = self.policy.resource0.resource.bucket
        return schema

    def get_debug_data(self, debug_sql):
        return self.extractor.run(debug_sql)

    def get_metrics_data(self):
        if self.policy.policy_type == "custom_check":
            metrics_sql = self._custom_query()
        elif self.policy.policy_type == "column":
            metrics_sql = self._metrics_query()
        elif self.policy.policy_type == "table":
            metrics_sql = self._table_query()
        elif self.policy.policy_type == "schema":
            # TODO: Make this better
            result = self._schema_policy()
            return result
        else:
            raise Exception("Unknown policy type")

        metrics = []
        for sql in metrics_sql:
            metric = self.extractor.run(sql)
            metrics.append(metric)

        return metrics



from enum import Enum


class TableMetricType(Enum):
    ROW_COUNT = "_row_count"
    DUPLICATE_COUNT = "_duplicate_count"
    NULL_COUNT_TABLE = "_null_count_table"
    PERCENTAGE_DISTINCT_COUNT = "_percentage_distinct_count"
    FRESHNESS = "_freshness"



from typing import List, Text

from altimate_models.shared_models import Policy, Resource
from altimate_profiler.builders.base_builder import Base
from altimate_profiler.metrics.table_metric_type import TableMetricType


class TableQueryBuilder(Base):
    def __init__(
        self,
        dialect,
        policy: Policy,
    ):
        self.policy = policy
        self.dialect = dialect

    def _table_sql(self, table, filters):
        if filters:
            table += " WHERE " + filters
        return table

    def _gen_sql(self, metric, params, table):
        attr = getattr(self.dialect, metric._value_)
        return attr(table, params)

    def _gen_table_path(self, resource: Resource):
        return "{}.{}.{}".format(
            resource.resource.database,
            resource.resource.db_schema,
            resource.resource.table,
        )

    def compile(self) -> List[Text]:
        metrics_queries = []
        all_filters = " AND ".join(
            x for x in [self.policy.resource0.filters, self.policy.filters] if x
        )
        table = self._table_sql(
            self._gen_table_path(self.policy.resource0), all_filters
        )
        for term in self.policy.terms:
            metrics_queries.append(
                self._gen_sql(TableMetricType[term.metric.upper()], term.params, table)
            )
        return metrics_queries



import os
from typing import Text, Dict


def gen_table_path(bucket: Text, path: Text, options: Dict, function: Text):
    options_string = ""
    file_path = f"s3://{os.path.join(bucket, path)}"
    table = f"{function}('{file_path}'"

    if options:
        options_string = ", " + ", ".join([f"{k}={v}" for k, v in options.items()])

    table += options_string + ")"

    return table



class TableauExtractor:
    def __init__(self, connection_str=None, data_source=None):
        self.connection_str = connection_str
        self.data_source = data_source

    def test(self):
        return self.data_source.test_connection()



from enum import Enum
from typing import Any, Dict, List, Optional, Text, Union

from pydantic.v1 import BaseModel, Field


# TODO: This needs to be removed in favor of Policy. Relevant parts in CLI also needs to be updated.
class TableMeta(BaseModel):
    database: Optional[Text]
    db_schema: Optional[Text] = Field(alias="schema")
    table: Optional[Text]


class Term(BaseModel):
    term_index: int
    metric: Text
    params: Dict[Text, Any]


class ResourceType(Enum):
    FILE = "file"
    TABLE = "table"


class TableResource(BaseModel):
    database: Text
    db_schema: Optional[Text] = Field(alias="schema", default=None)
    table: Optional[Text] = None


class S3Resource(BaseModel):
    bucket: Text
    path: Text
    format: Text
    options: Optional[Dict[Text, Any]]


class Resource(BaseModel):
    resource_type: ResourceType
    resource: Union[TableResource, S3Resource]
    filters: Optional[Text] = None


class Policy(BaseModel):
    policy_index: int
    policy_type: Text
    resources: List[Resource]
    filters: Optional[Text]
    terms: List[Term]

    @property
    def resource0(self):
        return self.resources[0]

    @property
    def resource1(self):
        return self.resources[1]


class ProfileConfig(BaseModel):
    policies: List[Policy]

class PrivateKeyType(Enum):
    PEM = "pem"
    DER = "der"
    


from .main import get_metrics_profile, get_schema_profile, get_debug_data, get_metadata
from .factory import ExtractorFactory

__all__ = [
    "get_metrics_profile",
    "get_schema_profile",
    "get_debug_data",
    "get_metadata",
    "ExtractorFactory",
]



from typing import List, Text

from altimate_models.shared_models import Policy
from altimate_profiler.builders.base_builder import Base
from altimate_profiler.metrics.column_metric_type import ColumnMetricType


class ColumnMetricsQueryBuilder(Base):
    def __init__(self, dialect, discovery_data, policy: Policy):
        self.dialect = dialect
        self.discovery_data = discovery_data
        self.policy = policy

    def _base_query_sample(self, select_sql: Text, filters: Text = None):
        sql = """
            SELECT 
                {select_sql}
            FROM {database}.{schema}.{table}
        """.format(
            select_sql=select_sql,
            table=self.policy.resource0.resource.table,
            schema=self.policy.resource0.resource.db_schema,
            database=self.policy.resource0.resource.database,
        )

        if filters:
            sql += f" WHERE {filters}"

        return sql

    def _extract_col_info(self, column):
        return column["COL_NAME"], column["COL_TYPE"]

    def _transform_col_info_metric(self, col_name, metric, table_name):
        alias = "{}__{}".format(col_name, metric._value_)
        attr = getattr(self.dialect, metric._value_)

        if not attr:
            raise Exception(
                "Unreachable: Metric type is defined that does not resolve to a definition."
            )

        select_unformatted = attr()
        select_no_alias = select_unformatted.format("{}".format(col_name))
        select = "{} AS {}".format(select_no_alias, alias)

        return select

    def _transform_col_info(self, col_name, col_type, table_name):
        terms = list(
            filter(lambda x: x.params.get("column") == col_name, self.policy.terms)
        )

        col_sql = [
            self._transform_col_info_metric(
                col_name, ColumnMetricType[term.metric.upper()], table_name
            )
            for term in terms
        ]
        if len(col_sql) == 0:
            return None

        return ",\n\t".join(col_sql)

    def _col_sql(self, item, select_body):
        table_name = item["NAME"]

        col_name, col_type = self._extract_col_info(item)
        col_sql = self._transform_col_info(col_name, col_type, table_name)

        if table_name not in select_body:
            select_body[table_name] = {"sql": [], "timestamp_fields": []}

        if col_sql is not None:
            select_body[table_name]["sql"].append(col_sql)
        # TODO: Better handling of date / time types is needed
        if col_type == "date" or "timestamp" in col_type:
            select_body[table_name]["timestamp_fields"].append(col_name)

    def _select_sql(self, ddata):
        select_body = {}

        [self._col_sql(item, select_body) for item in ddata]
        for table_name, cols in select_body.items():
            select_body[table_name] = {
                "sql": ",\n".join(cols["sql"]),
                "timestamp_fields": cols["timestamp_fields"],
            }

        return select_body

    def _timestamp_field(self, cols):  # TODO: Improve
        if len(cols["timestamp_fields"]) == 0:
            return None

        return cols["timestamp_fields"][0]

    def _gen_sql(self, metric, params, alias):
        attr = getattr(self.dialect, metric._value_)

        if not attr:
            raise Exception(
                "Unreachable: Metric type is defined that does not resolve to a definition."
            )

        select_unformatted = attr()
        select_no_alias = select_unformatted.format(**params)
        select = "{} AS {}".format(select_no_alias, alias)

        return select

    def compile(self) -> List[Text]:
        metrics_query = []

        for t_idx, term in enumerate(self.policy.terms):
            metrics_query.append(
                self._gen_sql(
                    ColumnMetricType[term.metric.upper()], term.params, f"term_{t_idx}"
                )
            )

        # aggregate different level of filters
        all_filters = " AND ".join(
            x for x in [self.policy.resource0.filters, self.policy.filters] if x
        )

        select_sql = ",\n".join(metrics_query)
        query = self._base_query_sample(select_sql=select_sql, filters=all_filters)

        return [query]



from typing import List, Text, Dict

from altimate_profiler.metrics.schema_metric_type import SchemaMetricType
from altimate_models.shared_models import Policy
from altimate_profiler.builders.base_builder import Base


class S3SchemaQueryBuilder(Base):

    def __init__(
            self,
            policy: Policy,
            discovery_data: dict,
    ):
        self.policy = policy
        self.discovery_data = discovery_data

    def _columns_present(self, params):
        columns = [column.lower() for column in params["columns"]]
        existing_columns = [
            row["column_name"].lower() for row in self.discovery_data["rows"]
        ]
        return all([column in existing_columns for column in columns])

    def _columns_not_present(self, params):
        columns = [column.lower() for column in params["columns"]]
        existing_columns = [
            row["column_name"].lower() for row in self.discovery_data["rows"]
        ]
        return all([(column not in existing_columns) for column in columns])

    def _has_primary_key(self, params):
        raise NotImplementedError("Not Applicable for this source")

    def _has_consistent_fk_names(self, params):
        raise NotImplementedError("Not Applicable for this source")

    def _index_with_pattern(self, params):
        raise NotImplementedError("Not Applicable for this source")

    def _column_data_type(self, params):
        column = params["column"]
        for row in self.discovery_data["rows"]:
            if row["column_name"].lower() == column.lower():
                return row["column_type"].lower()

    def _is_same_as(self, params):
        raise NotImplementedError("Not Implemented sama as check")

    def compile(self) -> List[Dict]:
        row = {}
        columns = []
        for idx, term in enumerate(self.policy.terms):
            column_name = term.metric + "_" + str(idx)
            columns.append(column_name)
            row[column_name] = getattr(self, SchemaMetricType[term.metric.upper()].value)(term.params)

        return [{
            "columns": columns,
            "rows": [row]
        }]


import sys
from contextlib import contextmanager
from typing import Any, Dict, Generator, Optional, Text

from pydantic.v1.main import BaseModel

from altimate_models.constants import DEFAULT_CLIENT_DB_CONNECTION_TIMEOUT

try:
    from sqlalchemy import create_engine
except ImportError:
    pass


class DataSource(BaseModel):
    id: Optional[Text]
    name: Text
    type: Text
    description: Optional[Text]
    connection_info: Optional[Text]
    connection_hash: Optional[Text]
    statement_timeout_in_ms: Optional[int]
    key_pair: Optional[Dict]
    auth_type: Optional[Text]

    @classmethod
    def prompt(cls, data_store_config: Dict):
        raise NotImplementedError()

    def cred_prompt(self):
        raise NotImplementedError()

    @classmethod
    def get_connection_information(cls, data_store_config: Dict):
        raise NotImplementedError()

    def get_databases(self):
        raise NotImplementedError()

    def test_connection(self):
        raise NotImplementedError()

    def get_connection_string(self, *args, **kwargs) -> Text:
        raise NotImplementedError

    def drop_credentials(self):
        raise NotImplementedError

    def get_resource(self):
        raise NotImplementedError

    def get_identifier_and_fqn(self):
        raise NotImplementedError

    @classmethod
    def get_resource_config(cls, resource_config: Dict):
        raise NotImplementedError()

    @classmethod
    def map_config_to_source(cls, config: Dict):
        raise NotImplementedError()

    @classmethod
    def get_config_from_url(cls, connection_info: str):
        raise NotImplementedError()

    def update_credentials(self, key: str):
        raise NotImplementedError()

    def override_schema(self, schema):
        raise NotImplementedError()

    def get_connect_args(self):
        return {"connect_timeout": DEFAULT_CLIENT_DB_CONNECTION_TIMEOUT}

    @contextmanager
    def get_engine(self):
        """
        has to be run as a context manager
        or else the tunnel will close
        """
        if not "sqlalchemy" in sys.modules:
            raise ImportError(
                "sqlalchemy is not installed. Please install sqlalchemy to test connections"
            )
        with self.get_bound_host_and_port() as conn_args:
            # connect_args might not be supported by all dialects
            engine = create_engine(
                self.get_connection_string(**conn_args),
                connect_args=self.get_connect_args(),
            )
            try:
                yield engine
                # the ssh tunnel would close up after the outer context manager exits
            except Exception as e:
                # re-raise the exceptions here so as to not catch them here and silently fail
                raise e
            finally:
                engine.raw_connection().close()
                if engine.dispose:
                    engine.dispose()

    @contextmanager
    def get_bound_host_and_port(self) -> Generator[Dict[str, Any], None, None]:
        # this is where the ssh tunneling would happen
        try:
            args: Dict[str, Any] = {}
            if hasattr(self, "host"):  # type: ignore
                args["host"] = self.host
            if hasattr(self, "port"):  # type: ignore
                args["port"] = self.port
            # we might have more args in the future. add them here
            yield args  # type: ignore
        except Exception as e:
            # re-raise exceptions raised in the context block
            raise e

    def get_dict(self):
        config = self.dict()
        config.pop("connection_hash")
        config.pop("connection_info")
        return config



from enum import Enum


class SchemaMetricType(Enum):
    COLUMNS_PRESENT = "_columns_present"
    COLUMNS_NOT_PRESENT = "_columns_not_present"
    HAS_PRIMARY_KEY = "_has_primary_key"
    HAS_CONSISTENT_FK_NAMES = "_has_consistent_fk_names"
    INDEX_WITH_PATTERN = "_index_with_pattern"
    COLUMN_DATA_TYPE = "_column_data_type"
    IS_SAME_AS = "_is_same_as"



import logging
from typing import Dict, Optional, Text

import requests
import tableauserverclient as TSC

from altimate_models.base.source import DataSource
from altimate_profiler.utils import sha256_encoding

logger = logging.getLogger(__name__)


class TableauSource(DataSource):
    """
    TableauSource represents a connection to a Tableau server.
    """

    host: Optional[str] = None
    site_name: Optional[str] = None
    pat_name: Optional[str] = None
    pat_secret: Optional[str] = None
    user: Optional[str] = None
    password: Optional[str] = None
    account: Optional[str] = None
    role: Optional[str] = None
    warehouse: Optional[str] = None

    @classmethod
    def prompt(cls, data_store_config: Dict):
        """Prompt for any additional configuration needed."""
        return data_store_config

    def cred_prompt(self):
        """Prompt for credentials if needed."""
        return None

    @classmethod
    def get_connection_information(cls, data_store_config: Dict):
        """Get connection information from the config."""
        data_store_config[
            "connection_info"
        ] = f"{data_store_config['host']}/{data_store_config['site_name']}"
        data_store_config["connection_hash"] = sha256_encoding(
            data_store_config["connection_info"]
        )
        return data_store_config

    def get_databases(self):
        """Get list of available databases (sites in Tableau context)."""
        if not (self.pat_name and self.pat_secret and self.host):
            return []

        try:
            # Authenticate with Tableau Server using PAT
            tableau_auth = TSC.PersonalAccessTokenAuth(
                token_name=self.pat_name,
                personal_access_token=self.pat_secret,
                site_id="",  # Empty string to list all sites
            )
            url = f"https://{self.host}"
            server = TSC.Server(url, use_server_version=True)

            with server.auth.sign_in(tableau_auth):
                # Get all available sites
                sites, _ = server.sites.get()
                # Return site content_urls as dictionaries with database_name key
                return [
                    {"database_name": site.content_url or "default"} for site in sites
                ]

        except Exception as e:
            # Log error and return empty list if connection fails
            logger.error(f"Error getting Tableau sites: {e}")
            return []

    @staticmethod
    def get_required_workbooks_and_sheets():
        """
        Get the required workbooks and their sheets for Tableau server validation.
        
        Returns:
            dict: Dictionary mapping workbook names to lists of required sheet names
        """
        return {
            'altimate_ts_events': ['access_events'],
            'altimate_ts_background_tasks': ['refresh_events'],
            'altimate_ts_users': ['user_data'],
        }

    def test_connection(self):
        """Test the connection to the Tableau server."""
        if self.pat_name and self.pat_secret:
            tableau_auth = TSC.PersonalAccessTokenAuth(
                token_name=self.pat_name,
                personal_access_token=self.pat_secret,
                site_id=self.site_name if self.site_name != "default" else "",
            )
            url = f"https://{self.host}"
            server = TSC.Server(url, use_server_version=True)
            with server.auth.sign_in(tableau_auth):
                # Test basic connection
                connection_test = True
                
                # Test for specific workbooks and sheets
                workbook_test = self._test_required_workbooks(server)
                
                return connection_test and workbook_test
        else:
            raise Exception(
                "failed to connect to Tableau server, missing PAT credentials"
            )

    def _test_required_workbooks(self, server):
        """Test for required workbooks and sheets using metadata API."""
        try:
            # Get the auth token from the signed-in session
            auth_token = server.auth_token
            
            # GraphQL query to get workbooks and their sheets
            query = """
            {
                workbooks {
                    name
                    sheets {
                        name
                    }
                }
            }
            """
            
            # Make request to metadata API
            metadata_url = f"https://{self.host}/api/metadata/graphql"
            headers = {
                'X-Tableau-Auth': auth_token,
                'Content-Type': 'application/json'
            }
            
            payload = {
                "query": query,
                "variables": {}
            }
            
            response = requests.post(metadata_url, json=payload, headers=headers)
            
            if response.status_code != 200:
                logger.error(f"Metadata API request failed with status {response.status_code}")
                return False
            
            data = response.json()
            
            if 'errors' in data:
                logger.error(f"GraphQL errors: {data['errors']}")
                return False
            
            workbooks = data.get('data', {}).get('workbooks', [])
            
            # Get required workbooks and sheets from static method
            required_workbooks = self.get_required_workbooks_and_sheets()
            
            # Check for required workbooks and sheets
            found_workbooks = {}
            for workbook in workbooks:
                workbook_name = workbook.get('name', '')
                if workbook_name in required_workbooks:
                    sheet_names = [sheet.get('name', '') for sheet in workbook.get('sheets', [])]
                    found_workbooks[workbook_name] = sheet_names
            
            # Validate requirements
            missing_requirements = []
            for required_workbook, required_sheets in required_workbooks.items():
                if required_workbook not in found_workbooks:
                    missing_requirements.append(f"Workbook '{required_workbook}' not found")
                else:
                    found_sheets = found_workbooks[required_workbook]
                    for required_sheet in required_sheets:
                        if required_sheet not in found_sheets:
                            missing_requirements.append(
                                f"Sheet '{required_sheet}' not found in workbook '{required_workbook}'"
                            )
            
            if missing_requirements:
                logger.warning(f"Missing requirements: {missing_requirements}")
                return False
            
            logger.info("All required workbooks and sheets found successfully")
            return True
            
        except Exception as e:
            logger.error(f"Error testing required workbooks: {e}")
            return False

    def get_connection_string(self, *args, **kwargs) -> Text:
        """Get connection string for Tableau."""
        return f"tableau://{self.host}/{self.site_name}"

    def drop_credentials(self):
        """Clear sensitive credentials."""
        config_dict = self.dict()
        config_dict.pop("password")
        config_dict.pop("user")
        return config_dict

    def get_resource(self):
        """Get resource information."""
        return {"url": self.host, "site": self.site_name}

    def get_identifier_and_fqn(self):
        """Get identifier and fully qualified name."""
        return (
            f"tableau://{self.host}/{self.site_name}",
            f"{self.host}/{self.site_name}",
        )

    @classmethod
    def get_resource_config(cls, resource_config: Dict):
        """Get resource configuration."""
        return resource_config

    @classmethod
    def map_config_to_source(cls, config: Dict):
        """Map configuration to source attributes."""
        connection_info = config.get("connection_info")
        if not connection_info:
            return config
        tableau_config = cls.get_config_from_url(connection_info)
        config.update(tableau_config)
        config.pop("connection_info", None)
        return config

    @classmethod
    def get_config_from_url(cls, connection_info: str):
        parts = connection_info.split("/", 1)
        if len(parts) != 2:
            raise ValueError("Invalid Tableau connection string format")
        return {"host": parts[0], "site_name": parts[1]}

    def update_credentials(self, key: str):
        """Update credentials using the provided key."""
        if ":" in key:
            pat_name, pat_secret = key.split(":", 1)
            self.pat_name = pat_name
            self.pat_secret = pat_secret
            self.user = pat_name
            self.password = pat_secret

    def override_schema(self, schema):
        """Override schema if needed."""
        return schema

    def get_connect_args(self):
        """Get connection arguments."""
        return super().get_connect_args()



import logging
from typing import Optional, Text

from sqlalchemy import create_engine, inspect

from altimate_models.base.source import DataSource
from altimate_models.constants import TEST_SQL_QUERY


class SQLAlchemyExtractor:
    # TODO: currently allowing connection_str as well as datasource.
    # should get rid of raw strings passed around at some point: AL-639
    def __init__(
        self,
        connection_str: Optional[Text] = None,
        data_source: Optional[DataSource] = None,
    ):
        self.connection_str = connection_str
        self.data_source: DataSource = data_source
        self.engine = None
        self.connection = None

    def _initialize(self):
        if self.engine and self.connection:
            return

        try:
            self.engine = create_engine(self.connection_str)
            self.connection = self.engine.connect()
        except Exception:
            raise

    def _execute(self, sql: Text):
        if not self.connection:
            raise Exception("Connection has already been closed. Could not execute.")

        cs = self.connection.execute(sql)
        results = self._retrieve_results(cs)

        return results

    def _retrieve_results(self, cs):
        columns = [d.name for d in cs.cursor.description]
        rows = [dict(zip(columns, row)) for row in cs.fetchall()]

        return {
            "columns": columns,
            "rows": rows,
        }

    def _terminate(self):
        if self.connection is not None:
            self.connection.close()
            self.connection = None

        if self.engine is not None:
            self.engine.dispose()
            self.engine = None

    def test(self):
        try:
            if self.data_source:
                with self.data_source.get_engine() as engine:
                    with engine.connect() as connection:
                        self.connection = connection
                        result = self._execute(TEST_SQL_QUERY)
                        rows = result["rows"]
                        columns = result["columns"]

                        return_result = (
                            len(rows) == 1 and rows[0][columns[0]] == 1
                        )  # TODO: Fix Indexing

                return return_result

            else:
                # Keeping this to support older connection_str method
                self._initialize()
                result = self._execute(TEST_SQL_QUERY)
                self._terminate()

                rows = result["rows"]
                columns = result["columns"]

                return len(rows) == 1 and rows[0][columns[0]] == 1  # TODO: Fix Indexing
        except Exception as e:
            logging.error(e)
            return False

    def foreign_keys(self, table_name: Text):
        self._initialize()

        inspector = inspect(self.engine)
        foreign_keys = inspector.get_foreign_keys(table_name)

        return foreign_keys

    def run(self, sql: Text):
        if self.data_source:
            with self.data_source.get_engine() as engine:
                with engine.connect() as connection:
                    self.connection = connection
                    results = self._execute(sql)
        else:
            self._initialize()
            results = self._execute(sql)
            self._terminate()

        return results



"""
Converting the Profiler output to that that is expected by the contract definition
TODO: Crate an abstraction which takes care of this mapping. Or consider doing it in the profiler.
"""
from typing import List, Dict


def generate_metadata(info: Dict):
    if "database" in info:
        return {
            "identifier": f"{info['database']}.{info['schema']}.{info['table']}"
        }
    elif "BUCKET" in info:
        return {
            "identifier": f"{info['BUCKET']}/{info['FILE_PATH']}?file_format={info['FILE_FORMAT']}",
            "bucket": info['BUCKET']
        }
    else:
        raise Exception("Invalid metadata")


def schema_mapper(schemas) -> List[Dict]:
    """
    Function which maps the schema generated by the profiler to the format
    expected by the contract definition
    :param columns_info: The schema generated by the profiler
    return: The schema in the format expected by the contract definition
    example:
    columns_info = [
        {
            "COL_NAME": "user_id",
            "COL_TYPE": "number",
            "COL_SORT_ORDER": "1",
            "DESCRIPTION": null,
            "IS_VIEW": "false",
            "table": "user",
            "schema": "dbt",
            "database": "analytics",
            "description": null,
            "id": "a3a617d01fba40b3abea57af0ee009a3"
        },...
    ]
    output: [
        {
            "metadata": {
                "database": "analytics",
                "schema": "dbt",
                "table": "user",
            },
            "table_schema": {
                 "column_name": "column_type"
            }
    ]
    # TODO: Create a class with the schemas defined.
    """
    resources = {}

    # Iterate over the column information
    for schema in schemas:
        for row in schema:
            resource = generate_metadata(row)
            identifier = resource["identifier"]
            column = {row["COL_NAME"]: row["COL_TYPE"]}
            if identifier in resources:
                resources[identifier]["resource_schema"].update(column)
            else:
                resources[identifier] = {"metadata": resource, "resource_schema": column}

    return list(resources.values())




from typing import List, Text

from altimate_models.shared_models import Policy, TableResource
from altimate_profiler.builders.column_metrics_query_builder import (
    ColumnMetricsQueryBuilder,
)
from altimate_profiler.builders.table_query_builder import TableQueryBuilder


class SQLAlchemySourceDialect:
    @classmethod
    def _row_count(cls, table, params):
        return "select COALESCE(COUNT(*), 0)  from {table}".format(table=table)

    @classmethod
    def _duplicate_count(cls, table, params):
        return "select COALESCE((SUM(c) - COUNT(c)), 0) as dulicates from (select md5(row_to_json(ROW({columns}))::TEXT),count(*) as c   from {table} group by md5(row_to_json(ROW({columns}))::TEXT) having count(*) > 1) as t".format(
            columns=",".join(params["columns"]), table=table
        )

    @classmethod
    def _numeric_variance(cls):
        return "variance({column})"

    @classmethod
    def _numeric_median(cls):
        raise NotImplementedError("Median is not implemented")

    @classmethod
    def _null_count_table(cls, table, params):
        raise NotImplementedError("null_count has not been implemented")

    @classmethod
    def _numeric_mode(cls):
        raise NotImplementedError("mode has not been implemented")

    @classmethod
    def _percentile(cls):
        raise NotImplementedError("percentile has not been implemented")

    @classmethod
    def _null_count(cls, table, params):
        raise NotImplementedError("null_count has not been implemented")

    @classmethod
    def _like_percentage(cls):
        raise NotImplementedError("like has not been implemented")

    @classmethod
    def _regex_match_percentage(cls):
        raise NotImplementedError("null_count has not been implemented")

    @classmethod
    def create_table_path(cls, resource: TableResource):
        return f"{resource.database}.{resource.db_schema}.{resource.table}"

    @classmethod
    def _percentage_distinct_count(cls, table, params):
        concat_columns = (
            "CONCAT("
            + ",".join(
                [
                    f"coalesce({column}::text, '')"
                    for column in params.get("columns", [])
                ]
            )
            + ")"
        )
        sql = """
        WITH total_rows AS (
          SELECT COUNT(*) AS total
          FROM {table}
        ),
        distinct_rows AS (
          SELECT COUNT(DISTINCT {concat_columns}) AS num_distinct
          FROM {table}
        )
        SELECT
          ROUND((num_distinct * 100.0) / total, 2) AS percentage_distinct_rows
        FROM total_rows, distinct_rows;
        
        """
        return sql.format(concat_columns=concat_columns, table=table)

    @classmethod
    def _approx_distinct_count(cls):
        return "COUNT(DISTINCT {column})"

    @classmethod
    def _approx_distinctness(cls):
        return "{} / CAST(COUNT(*) AS NUMERIC)".format(cls._approx_distinct_count())

    @classmethod
    def _freshness(cls):
        raise NotImplementedError("freshness has not been implemented")

    @classmethod
    def _numeric_mean(cls):
        return "AVG({column})"

    @classmethod
    def _numeric_min(cls):
        return "MIN({column})"

    @classmethod
    def _numeric_max(cls):
        return "MAX({column})"

    @classmethod
    def _numeric_std(cls):
        return "STDDEV(CAST({column} as double))"

    @classmethod
    def _allowed_values(cls):
        raise NotImplementedError("allowed_values has not been implemented")

    @classmethod
    def _dates_in_future_percentage(cls):
        raise NotImplementedError("dates_in_future_percentage has not been implemented")

    @classmethod
    def _cardinality(cls):
        return "COUNT(DISTINCT {column})"

    @classmethod
    def _allowed_values(cls):
        raise NotImplementedError("allowed_values has not been implemented")

    @classmethod
    def _mean_length(cls):
        return "AVG(LENGTH({column}))"

    @classmethod
    def _max_length(cls):
        return "MAX(LENGTH({column}))"

    @classmethod
    def _min_length(cls):
        return "MIN(LENGTH({column}))"

    @classmethod
    def _std_length(cls):
        return "STDDEV(CAST(LENGTH({column}) as double))"

    @classmethod
    def _categories(cls):
        return "ARRAY_AGG(DISTINCT CAST({column} AS varchar))"

    @classmethod
    def _num_categories(cls):
        return "COUNT(DISTINCT CAST({column} AS varchar))"

    @classmethod
    def _conditional_percentage(cls):
        raise NotImplementedError("conditional_percentage has not been implemented")

    @classmethod
    def _conditional_max(cls):
        raise NotImplementedError

    @classmethod
    def _conditional_min(cls):
        raise NotImplementedError

    @classmethod
    def _conditional_mean(cls):
        raise NotImplementedError

    @classmethod
    def _conditional_count(cls):
        raise NotImplementedError

    @classmethod
    def _text_int_rate(cls):
        raise NotImplementedError

    @classmethod
    def _text_number_rate(cls):
        raise NotImplementedError

    @classmethod
    def _text_uuid_rate(cls):
        raise NotImplementedError

    @classmethod
    def text_all_spaces_rate(cls):
        raise NotImplementedError

    @classmethod
    def _text_null_keyword_rate(cls):
        raise NotImplementedError

    @classmethod
    def _zero_rate(cls):
        raise NotImplementedError

    @classmethod
    def _negative_rate(cls):
        raise NotImplementedError

    @classmethod
    def _completeness(cls):
        raise NotImplementedError

    @classmethod
    def schema_tables_query(
        cls,
        database_name: Text,
        schema_name: Text,
        table_name: Text = None,
        filters: Text = None,
    ):
        raise NotImplementedError

    @classmethod
    def schema_columns_query(
        cls,
        database_name: Text,
        schema_name: Text,
        table_name: Text = None,
        filters: Text = None,
        schemas_to_exclude: List[Text] = None,
        use_account_usage: bool = False,
    ):
        raise NotImplementedError

    @classmethod
    def column_metrics_query(cls, discovery_data, policy: Policy) -> List[Text]:
        builder = ColumnMetricsQueryBuilder(cls, discovery_data, policy)
        query = builder.compile()
        return query

    @classmethod
    def table_metrics_query(cls, policy: Policy) -> List[Text]:
        builder = TableQueryBuilder(cls, policy)
        query = builder.compile()
        return query

    @classmethod
    def access_logs_query(cls):
        raise NotImplementedError

    @classmethod
    def copy_and_load_logs_query(cls):
        raise NotImplementedError



from typing import Text, Dict

from altimate_models.duckdb.dialect import DuckDBSourceDialect
from altimate_models.s3.utils import gen_table_path
from altimate_models.shared_models import S3Resource


class JsonSourceDialect(DuckDBSourceDialect):

    DATA_QUERY = "data"
    METADATA_QUERY = "metadata"

    FORMAT_TO_TABLE_FUNCTION = {
        DATA_QUERY:  "read_json",
        METADATA_QUERY: "read_json",
    }

    @classmethod
    def create_table_path(cls, resource: S3Resource, query_type: Text) -> Text:
        return gen_table_path(resource.bucket, resource.path, resource.options, cls.FORMAT_TO_TABLE_FUNCTION[query_type])

    @classmethod
    def get_options(cls, options: Dict):
        if options is None:
            options = {}
        if not options.get('columns'):
            options['auto_detect'] = True
        return options

    @classmethod
    def schema_tables_query(
            cls,
            bucket: Text,
            path: Text,
            options: Dict
    ):
        options = cls.get_options(options)
        options["json_format"] = "array_of_records"
        return f" DESCRIBE SELECT * FROM " \
               f"{gen_table_path(bucket, path, options, cls.FORMAT_TO_TABLE_FUNCTION[cls.METADATA_QUERY])}"

    @classmethod
    def _base_query_sample(
            cls,
            select_sql: Text,
            bucket: Text,
            path: Text,
            options: Dict,
            filters: Text = None
    ):
        options = cls.get_options(options)
        table = gen_table_path(bucket, path, options, cls.FORMAT_TO_TABLE_FUNCTION[cls.DATA_QUERY])
        sql = """
            SELECT 
                {select_sql}
            FROM {table}
        """.format(
            select_sql=select_sql,
            table=table,
        )
        if filters:
            sql += f" WHERE {filters}"
        return sql

    @classmethod
    def _table_sql(
            self,
            bucket: Text,
            path: Text,
            options: Dict,
            filters=None
    ):
        options = self.get_options(options)
        table = gen_table_path(bucket, path, options, self.FORMAT_TO_TABLE_FUNCTION[self.DATA_QUERY])
        if filters:
            table += " WHERE " + filters
        return table



from altimate_models.duckdb.extractor import DuckDBExtractor


class S3Extractor(DuckDBExtractor):

    def _initialize(self):
        if self.engine:
            return
        try:
            self.engine = self._create_engine(self.connection_str)
        except Exception:
            raise Exception("Could not connect to S3 with the provided credentials.")


from typing import List, Text

from altimate_models.base.dialect import SQLAlchemySourceDialect
from altimate_profiler.exceptions import AltimateProfilerException


class PostgreSQLSourceDialect(SQLAlchemySourceDialect):
    TIME_UNIT_TO_SECOND_MAPPING = {
        "second": 1,
        "minute": 60,
        "hour": 3600,
        "day": 86400,
        "week": 604800,
        "month": 2592000,
        "year": 31536000,
    }

    @classmethod
    def _freshness(cls, table, params):
        column = params["column"]
        time_unit = params.get("timeUnit", "day")

        divide_by = cls.TIME_UNIT_TO_SECOND_MAPPING.get(time_unit)

        if not divide_by:
            raise AltimateProfilerException(
                "Invalid timeUnit. It must be one of {}".format(
                    ", ".join(cls.TIME_UNIT_TO_SECOND_MAPPING.keys())
                )
            )

        return (
            "SELECT MIN(EXTRACT(EPOCH FROM (NOW() - {column}))) "
            "/ {divide_by} AS freshness FROM {table}".format(
                column=column, divide_by=divide_by, table=table
            )
        )

    @classmethod
    def _numeric_std(cls):
        return "STDDEV(CAST({column} as double precision))"

    @classmethod
    def _text_int_rate(cls):
        return (
            "SUM(CASE WHEN CAST({column} AS varchar) ~ '^([-+]?[0-9]+)$' THEN 1 ELSE 0 END) /"
            " COALESCE(COUNT(*), 1)"
        )

    @classmethod
    def _null_percentage(cls):
        return "SUM(CASE WHEN {column} IS NULL THEN 1 ELSE 0 END) * 100/COALESCE(COUNT(*), 1)"

    @classmethod
    def _text_number_rate(cls):
        return (
            "SUM(CASE WHEN CAST({column} AS varchar) ~"
            " '^([-+]?[0-9]*[.]?[0-9]+([eE][-+]?[0-9]+)?)$' THEN 1 ELSE 0 END) /"
            " COALESCE(COUNT(*), 1)"
        )

    @classmethod
    def _text_uuid_rate(cls):
        return (
            "SUM(CASE WHEN CAST({column} AS varchar) ~"
            " '^([0-9a-fA-F]{{8}}-[0-9a-fA-F]{{4}}-[0-9a-fA-F]{{4}}-[0-9a-fA-F]{{4}}-[0-9a-fA-F]{{12}})$'"
            " THEN 1 ELSE 0 END) / COALESCE(COUNT(*), 1)"
        )

    @classmethod
    def _invalid_email_rate(cls):
        return (
            "SUM(CASE WHEN CAST({column} AS varchar) ~"
            " '^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+.[a-zA-Z0-9\\\-.]+$'"
            " THEN 0 ELSE 1 END) / COALESCE(COUNT(*), 1)"
        )

    @classmethod
    def _allowed_values(cls):
        return "SUM(CASE WHEN {column} NOT IN {values} THEN 1 ELSE 0 END)"

    @classmethod
    def _text_all_spaces_rate(cls):
        return (
            "SUM(CASE WHEN CAST({column} AS varchar) ~ '^(\\\\s+)$' THEN 1 ELSE 0 END) /"
            " COALESCE(COUNT(*), 1)"
        )

    @classmethod
    def _text_null_keyword_rate(cls):
        return (
            "SUM(CASE WHEN UPPER(CAST({column} as varchar)) IN ('NULL', 'NONE', 'NIL',"
            " 'NOTHING') THEN 1 ELSE 0 END) / COALESCE(COUNT(*), 1)"
        )

    @classmethod
    def _zero_rate(cls):  # TODO: ?
        return (
            "SUM(CASE WHEN UPPER(CAST({column} as varchar)) IN ('NULL', 'NONE', 'NIL',"
            " 'NOTHING') THEN 1 ELSE 0 END) / COALESCE(COUNT(*), 1)"
        )

    @classmethod
    def _negative_rate(cls):
        return "SUM(CASE WHEN {column} < 0 THEN 1 ELSE 0 END) / COALESCE(COUNT(*), 1)"

    @classmethod
    def _completeness(cls):
        return "COUNT({column}) / COALESCE(COUNT(*), 1)"

    @classmethod
    def _dates_in_future_percentage(cls):
        return (
            "SUM(CASE WHEN {column} > CURRENT_TIMESTAMP THEN 1 ELSE 0 END) * 100"
            " / COALESCE(COUNT(*), 1) "
        )

    @classmethod
    def _conditional_max(cls):
        return "MAX(CASE WHEN {condition} THEN {column} ELSE NULL END)"

    @classmethod
    def _conditional_min(cls):
        return "MIN(CASE WHEN {condition} THEN {column} ELSE NULL END)"

    @classmethod
    def _conditional_count(cls):
        return "SUM(CASE WHEN {condition} THEN 1 ELSE 0 END)"

    @classmethod
    def _conditional_percentage(cls):
        return (
            "SUM(CASE WHEN {condition} THEN 1 ELSE 0 END) * 100 / COALESCE(COUNT(*), 1)"
        )

    @classmethod
    def _conditional_mean(cls):
        return "AVG(CASE WHEN {condition} THEN {column} ELSE 0 END)"

    @classmethod
    def _null_count(cls):
        return "SUM(CASE WHEN {column} IS NULL THEN 1 ELSE 0 END)"

    @classmethod
    def _regex_match_percentage(cls):
        return "SUM(CASE WHEN {column}::TEXT ~ '{pattern}' THEN 1.0 ELSE 0.0 END) * 100 /  COALESCE(COUNT(*), 1) "

    @classmethod
    def _like_percentage(cls):
        return "SUM(CASE WHEN {column}:: TEXT  LIKE '{pattern}' THEN 1.0 ELSE 0.0 END) * 100 / COALESCE(COUNT(*), 1)"

    @classmethod
    def _numeric_mode(cls):
        return "MODE() WITHIN GROUP (ORDER BY {column})"

    @classmethod
    def _percentile(cls):
        return "PERCENTILE_CONT({value}) WITHIN GROUP (ORDER BY {column})"

    @classmethod
    def _numeric_median(cls):
        return "PERCENTILE_CONT(0.5) WITHIN GROUP (ORDER BY {column})"

    @classmethod
    def _null_count_table(cls, table, params):
        condition = " AND ".join([f"{column} IS NULL" for column in params["columns"]])
        return (
            f"SELECT SUM(CASE WHEN {condition} THEN 1 ELSE 0 END) FROM {table}".format(
                table=table, condition=condition
            )
        )

    @classmethod
    def schema_tables_query(
        cls,
        database_name: Text,
        schema_name: Text,
        table_name: Text = None,
        filters: Text = None,
    ):
        raise NotImplementedError

    @classmethod
    def schema_columns_query(
        cls,
        database_name: Text,
        schema_name: Text,
        table_name: Text = None,
        filters: Text = None,
        schemas_to_exclude: List[Text] = ["information_schema", "pg_catalog"],
    ):
        sql = """
            SELECT lower(c.table_name) AS "NAME",
                lower(c.column_name) AS "COL_NAME",
                lower(c.data_type) AS "COL_TYPE",
                c.ordinal_position AS "COL_SORT_ORDER",
                lower(c.table_catalog) AS "DATABASE",
                lower(c.table_schema) AS "SCHEMA",
                CASE
                    lower(t.table_type)
                    WHEN 'view' THEN 'true'
                    ELSE 'false'
                END "IS_VIEW"
            FROM INFORMATION_SCHEMA.COLUMNS AS c
                LEFT JOIN INFORMATION_SCHEMA.TABLES t ON c.TABLE_NAME = t.TABLE_NAME
                AND c.TABLE_SCHEMA = t.TABLE_SCHEMA
            WHERE LOWER(c.table_catalog) = LOWER('{database_name}')
        """.format(
            database_name=database_name
        )
        excluded_schemas_string = ", ".join([f"'{s}'" for s in schemas_to_exclude])
        if table_name:
            sql += f" AND LOWER(c.table_name)=LOWER('{table_name}')"

        if schema_name:
            sql += f" AND LOWER(c.table_schema)=LOWER('{schema_name}')"
        else:
            sql += f" AND LOWER(c.table_schema) NOT IN ({excluded_schemas_string})"

        if filters:
            sql += f" AND {filters}"

        return sql

    @classmethod
    def query_access_logs_query(cls):
        raise NotImplementedError

    @classmethod
    def query_copy_logs_query(cls):
        raise NotImplementedError
